/*****************************************************************************
 *
 * File ..................: mbstuff.c (originally filesub.c)
 * Purpose ...............: Routines based on the mbse source. Some newly
 *                          written routines too.
 * Last modification date : 20-Oct-2000
 * Revised by ............: Johannes Lundberg, 2:206/149
 *
 *****************************************************************************
 * Copyright (C) 1997-2000
 *   
 * Michiel Broek		FIDO:		2:280/2802
 * Beekmansbos 10
 * 1971 BV IJmuiden
 * the Netherlands
 *
 * This BBS is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * MBSE BBS is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to the
 * Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************/

#include "fbutil.h"


struct fileareas area;
struct fileareashdr areahdr;
struct FILERecord file;

unsigned long number_of_fileareas;

struct _archiverhdr archiverhdr;
struct _archiver archiver;

/*
 * Open FileDataBase.
 */
FILE *OpenFileBase(unsigned long Area, int Write)
{
	FILE    *pFile;
	char    FileBase[81];
    
	sprintf(FileBase,"%s/fdb/fdb%ld.data", getenv("MBSE_ROOT"), Area);
                          
	if (Write)
		pFile = fopen(FileBase, "r+");
	else	
		pFile = fopen(FileBase, "r");
	
	if (pFile == NULL) {
		printf ("! Can't open file %s\n\n", FileBase);
	}
	
	return pFile;
}


/*
 * Open the fareas.data file for read or R/W and read the headerrecord.
 * The filepointer is at the start of the first record.
 */
FILE *OpenFareas(int Write)
{
	FILE    *pAreas;
	char    FileArea[81];

	sprintf(FileArea, "%s/etc/fareas.data", getenv("MBSE_ROOT"));

	if (Write)
		pAreas = fopen(FileArea, "r+");
	else
		pAreas = fopen(FileArea, "r");

	if (pAreas == NULL) {
		printf ("! Can't open areas database %s.\n", FileArea);
	} else
		fread(&areahdr, sizeof(areahdr), 1, pAreas);

	return pAreas;
}                                                                                                                                                                                   

/*
 * Make any string lower case.
 */
void lower_string (char *string)
{
  int index;
  
  for (index = 0; index < strlen (string); index++)
  {
    string[index] = tolower (string[index]);
  }
}

/*
 * Check if a file exists in a given directory. Case insensitive.
 */
int search_dir (char *filename, char *directory)
{
  DIR *dir;
  struct dirent *de;
  char name_lower[54];
  
  if ((dir = opendir (directory)) == NULL)
  {
    return -1;
  }

  lower_string (filename);
  
  while ((de = readdir (dir)))
  {
    strncpy (name_lower, de->d_name, 54);
    lower_string (name_lower);
    if (strncmp (name_lower, filename, 54) == 0)
    {
      strncpy (filename, de->d_name, 54);
      closedir (dir);
      return 0;
    }
  }

  closedir (dir);
  
  return 1; 
}

/*
 * Add file to the FileDataBase. If fileid is true, then try to
 * get the filedescription from FILE_ID.DIZ if it is in the
 * archive, else the user must supply the description.
 * Returns TRUE is successfull.
 */
int Addfile(char *File, int AreaNum, int fileid)
{
	FILE	*id, *pFileDB;
	int	iDesc = 1, GotId = FALSE;
	char	*Filename, *temp1;
	char	*Desc[26]; 
	struct	stat statfile; 
	int	i;
	char	temp[81];
	char	*arc;

	char	dizfile[54] = "file_id.diz";

	arc	= calloc(81, sizeof(char));

	Filename = calloc(81, sizeof(char));
	temp1    = calloc(257, sizeof(char));  

	FileAreaGetData(&area, AreaNum);

	
	sprintf(Filename, "%s/%s", area.Path, File);

	if ((pFileDB = OpenFileBase(AreaNum, TRUE)) != NULL) {
		/*
		 * Do a physical check of file to see if it exists
		 * if it fails it will return a zero which will not
		 * increase his uploads stats
		 */
		if(stat(Filename, &statfile) != 0) {

			printf("File %s does not exist.\n", Filename);

			fclose(pFileDB);
			free(Filename);
			free(temp1);
			return FALSE;
		}

		memset(&file, 0, sizeof(file));
		strcpy(file.Name, File);
		sprintf(temp1,"%ld",statfile.st_size);
		file.Size = atoi(temp1);
		file.FileDate = statfile.st_mtime;
		time (&file.UploadDate);

		printf ("Use default uploader (%s) [Y/n] ", DEFAULT_UPLOADER);
		if (tolower (getchar ()) == 'n')
		{
			printf ("Uploader: ");
			GetstrC (file.Uploader, 20);
		}
		else
		{
			strcpy (file.Uploader, DEFAULT_UPLOADER);
		}

		ArchiversInit ();

		if ((arc = GetFileType(File)) == NULL) {
			/*
			 * If the filetype is unknown, it is probably
			 * a textfile or so. Import it direct.
			 */
			printf ("- Unknown file type. Wouldn't be able to extract file_id.diz.\n");
			// ArchiverGetByFileType (&archiver, arc);
		} else {
			if (fileid) {
				ArchiverGetByFileType (&archiver, arc);
				sprintf(temp, "%s %s/%s file_id.diz >/dev/null",
				        archiver.iunarc, area.Path, File);
				printf ("  Extracting FILE_ID.DIZ\n");
				system (temp);

				if (search_dir (dizfile, "."))
				{
					/* The filetype is known, but the archive does
					   not unpack. Bad archive? */
					printf ("- No FILE_ID.DIZ found in %s\n", File);
				} else {
					GotId = TRUE;
					/* Found FILE_ID.DIZ in */
					printf("+ FILE_ID.DIZ found in %s\n", File);
					fflush(stdout);
				}
			}
		}

		if (GotId) {
			if ((id = fopen(dizfile, "r")) != NULL) {
				/*
				 * Import FILE_ID.DIZ, format to max. 25
				 * lines, 48 chars width.
				 */
				while ((fgets(temp1, 256, id)) != NULL) {
					if (iDesc < 26) {
						Striplf(temp1);
						temp1[48] = '\0';
						strcpy(file.Desc[iDesc - 1], temp1);
					}
					iDesc++;
				}
			}
			fclose(id);
			unlink(dizfile);
		} else {
			/*
			 * Ask the user for a description.
			 */
			for (i = 0; i < 26; i++)
				*(Desc + i) = (char *) calloc(49, sizeof(char));

 			printf("Please enter a description for %s.\n", File);
			while (TRUE) {
				printf("%2d> ", iDesc);
				fflush(stdout);

				GetstrC(*(Desc + iDesc), 48);

				if((strcmp(*(Desc + iDesc), "")) == 0) 
					break;

				iDesc++;

				if(iDesc >= 26)
					break;
			}

			for(i = 1; i < iDesc; i++)
				strcpy(file.Desc[i - 1], Desc[i]);

			for (i = 0; i < 26; i++)
				free(Desc[i]);
		}

		fseek(pFileDB, 0, SEEK_END);
		fwrite(&file, sizeof(file), 1, pFileDB);
		fclose(pFileDB);

#if 0
		sprintf(temp, "%s/log/uploads.log", getenv("MBSE_ROOT"));
		if ((pPrivate = fopen(temp, "a+")) == NULL)
//			WriteError("$Can't open %s", temp);
			printf ("Can't open %s.\n", temp);
		else {
			iPrivate = TRUE;
			fprintf(pPrivate, "****************************************************");
			fprintf(pPrivate, "\nUser        : %s", file.Uploader);
			fprintf(pPrivate, "\nFile        : %s", file.Name);
			fprintf(pPrivate, "\nSize        : %lu", file.Size);
			fprintf(pPrivate, "\nUpload Date : %s\n\n", StrDateDMY(file.UploadDate));
				
			for(i = 0; i < iDesc - 1; i++)
				fprintf(pPrivate, "%2d: %s\n", i, file.Desc[i]);

			fclose(pPrivate);
		}

#endif

	}

	free(Filename);
	free(temp1);
	return TRUE;
}

/*
 * Add a new file area
 */
void AddFileArea (struct fileareas areadata, unsigned long area_nr)
{
	FILE *fareas, *newareas;
	unsigned long counter;
	struct fileareas tmpdata;
	char command[240];
	
	FileAreasInit ();
        fareas = OpenFareas (FALSE);
	
	newareas = fopen ("/tmp/fareas.data", "w");

	fwrite (&areahdr, areahdr.hdrsize, 1, newareas);

	for (counter = 1; counter <= number_of_fileareas; counter++)
	{
		if (counter == area_nr)
		{
			fwrite (&areadata, areahdr.recsize, 1, newareas);
		}

		fread (&tmpdata, areahdr.recsize, 1, fareas);
		fwrite (&tmpdata, areahdr.recsize, 1, newareas); 
	}

	fclose (newareas);
	
	for (counter = number_of_fileareas; counter >= area_nr; counter--)
	{
		/* FIXME: This is dangerous. */
		sprintf (command, "mv %s/fdb/fdb%ld.data %s/fdb/fdb%ld.data",
			 getenv ("MBSE_ROOT"), counter, getenv ("MBSE_ROOT"),
			 counter + 1);
		system (command);
	}

	/* Create an empty fdb for the newly created file area. */
	sprintf (command, "touch %s/fdb/fdb%ld.data", getenv ("MBSE_ROOT"),
	         area_nr);
	system (command);
	
	/* FIXME: Do this in a better/more secure way? */
	sprintf (command, "mv /tmp/fareas.data %s/etc/fareas.data",
	         getenv ("MBSE_ROOT"));
        system (command);
}

/*
 * Delete a file area
 */
void DeleteFileArea (unsigned long area_nr)
{
	FILE *fareas, *newareas;
	unsigned long counter;
	struct fileareas tmpdata;
	char command[240];
	
	FileAreasInit ();
        fareas = OpenFareas (FALSE);
	
	newareas = fopen ("/tmp/fareas.data", "w");

	fwrite (&areahdr, areahdr.hdrsize, 1, newareas);

	for (counter = 1; counter <= number_of_fileareas; counter++)
	{
		fread (&tmpdata, areahdr.recsize, 1, fareas);

		if (counter != area_nr)
		{
			fwrite (&tmpdata, areahdr.recsize, 1, newareas); 
		}
	}

	fclose (newareas);

	/* FIXME: Backup the old fdb before overwriting it. */
	
	for (counter = area_nr; counter <= number_of_fileareas; counter++)
	{
		sprintf (command, "mv %s/fdb/fdb%ld.data %s/fdb/fdb%ld.data",
			 getenv ("MBSE_ROOT"), counter, getenv ("MBSE_ROOT"),
			 counter - 1);
		system (command);
	}

	sprintf (command, "mv /tmp/fareas.data %s/etc/fareas.data",
	         getenv ("MBSE_ROOT"));
        system (command);
}

/*
 * Set file area number, set global area description and path.
 */

void SetFileArea(unsigned long AreaNum)
{
	FILE	*pArea;
	long	offset;

	memset(&area, 0, sizeof(area));

	if ((pArea = OpenFareas(FALSE)) == NULL)
		return;

	offset = areahdr.hdrsize + ((AreaNum - 1) * sizeof (struct fileareas));
	if (fseek(pArea, offset, 0) != 0) {
		printf ("! Error: Can't move pointer in file areas database.\n");
		return;
	}

	fread(&area, areahdr.recsize, 1, pArea);

	fclose(pArea);
}

/*
 * Get the Name of file area X.
 */

int FileAreaGetData (struct fileareas *data, unsigned long area_nr)
{
	FILE	*pArea;
	long	offset;

	memset(data, 0, sizeof(struct fileareas));

	if ((pArea = OpenFareas(FALSE)) == NULL)
		return 1;

	offset = areahdr.hdrsize + ((area_nr - 1) * areahdr.recsize);
	if (fseek(pArea, offset, 0) != 0) {
		printf ("Error: Can't move pointer in fileareas database.\n");
		return 1;
	}

	fread(data, areahdr.recsize, 1, pArea);

	fclose(pArea);
	
	return 0;
}

unsigned long FileAreaGetByPath (char *path)
{
	unsigned long area_nr;
	struct fileareas data;

	for (area_nr = 1; area_nr <= number_of_fileareas; area_nr++)
	{
		FileAreaGetData (&data, area_nr);
		
		if (strcmp (data.Path, path) == 0)
		{
		  return area_nr;
		}
	}
	
	return 0;
}

int FileAreasInit (void)
{
	FILE    *pAreas;
	char    FileArea[81];

	struct stat	statfile;

	sprintf(FileArea, "%s/etc/fareas.data", getenv("MBSE_ROOT"));

	pAreas = fopen(FileArea, "r");

	if (pAreas == NULL){
		printf ("Can't open FileBase %s.\n", FileArea);
		return 1;
	} else
		fread(&areahdr, sizeof(areahdr), 1, pAreas);

	stat (FileArea, &statfile);
	number_of_fileareas = (statfile.st_size - areahdr.hdrsize) / areahdr.recsize; 
	
	fclose (pAreas);
	
	return 0;
}

/*
 * Set the delete-flag for a given file.
 */
int FileDelete (char *filename, long Area)
{
  FILE *file;
  struct FILERecord data;
  int counter, records; 

  FileAreasInit ();
  
  if ((file = OpenFileBase (Area, TRUE)) == NULL)
  {
    return -1;
  }

  fseek (file, 0, SEEK_END);
  records = ftell (file) / sizeof (struct FILERecord);
  rewind (file);

  fread (&data, 1, sizeof (struct FILERecord), file);

  for (counter = 0; counter < records; counter++)
  {
    if (!strcmp (data.Name, filename))
    {
      printf ("Found file %s at record %d.\n", filename, counter);
      data.Deleted = TRUE;
      fseek (file, -(sizeof (struct FILERecord)), SEEK_CUR);
      fwrite (&data, 1, sizeof (struct FILERecord), file);
    }
  }
  
  fclose (file);
  
  return 0;
}

/*
 * Initialize the archivers.
 */

int ArchiversInit (void)
{
        FILE    *fp;
        char    temp[81];

        sprintf(temp, "%s/etc/archiver.data", getenv("MBSE_ROOT"));
        if ((fp = fopen(temp, "r")) == NULL)
        {
        	printf ("Could not open %s.\n", temp);
        	printf ("No archivers defined?\n");
                return 1;
	}
	
        fread(&archiverhdr, sizeof(archiverhdr), 1, fp);
        
        fclose(fp);
        
        return 0;
}

/*
 * Check if an archiver exists for the given file type.
 */
int ArchiverGetByFileType (struct _archiver *data, char *filetype)
{
	FILE    *fp;
//	int     err, Found = FALSE;
	char    temp[81];
	long	offset;
//	char    *cwd = NULL;
	long	archivers;
	struct stat	statfile;

	memset(data, 0, sizeof(struct _archiver));

        sprintf(temp, "%s/etc/archiver.data", getenv("MBSE_ROOT"));

	if (stat (temp, &statfile))
	{
		printf ("! File %s does not exist. No archivers defined?\n", temp);
		return 1;
	}
	archivers = (statfile.st_size - archiverhdr.hdrsize) / archiverhdr.recsize;

        if ((fp = fopen(temp, "r")) == NULL)
                return 1;
	        
	offset = archiverhdr.hdrsize;
	if (fseek(fp, offset, 0) != 0) {
		printf ("! Could not open %s. File corrupt?\n", temp);
		return 1;
	}

	while (fread(data, archiverhdr.recsize, 1, fp) == 1) {
		if ((strcmp(filetype, data->name) == 0) && (data->available)) {
			/* We have found our archiver. */
                        return 0;
		}
	}
        fclose(fp);

	/* An archiver for this file type does not exist. */

	memset(data, 0, sizeof(struct _archiver));

	return 1;
}

/*
 * Try to find out the type of uploaded file.
 * FIXME: Make this function use the 'file' utillity.
 */
char *GetFileType(char *fn)
{
        unsigned char   buf[8], dbuf[80];
        FILE            *fp;
        int             i;

        if ((fp = fopen(fn, "r")) == NULL) {
//                WriteError("$Can't open file %s", fn);
		printf ("! Could not open file %s\n", fn);
                return NULL;
        }

        if (fread(buf, 1, sizeof(buf), fp) != sizeof(buf)) {
//                WriteError("$Can't read head of file %s", fn);
		printf ("- Could not read head of file %s\n", fn);
                return NULL;
        }

        fclose(fp);
        dbuf[0] = '\0';

        for (i = 0; i < sizeof(buf); i++)
                if ((buf[i] >= ' ') && (buf[i] <= 127))
                        sprintf((char*)dbuf+strlen(dbuf), "  %c", buf[i]);
                else
                        sprintf((char*)dbuf+strlen(dbuf), " %02x", buf[i]);

//        Syslog('b', "file head: %s", dbuf);

        /*
         * Various expected uploads. Not that the standard MS-DOS archivers
         * must return the exact format, ie "ZIP" for PKZIP. These strings
         * are tested against the archivers database. Others that aren't
         * compressed files are not important, they just pop up in your
         * logfiles.
         */
        if (memcmp(buf, "PK\003\004", 4) == 0)          return (char *)"ZIP";
        if (*buf == 0x1a)                               return (char *)"ARC";
        if (memcmp(buf+2, "-l", 2) == 0)                return (char *)"LZH";
        if (memcmp(buf, "ZOO", 3) == 0)                 return (char *)"ZOO";
        if (memcmp(buf, "`\352", 2) == 0)               return (char *)"ARJ";
        if (memcmp(buf, "Rar!", 4) == 0)                return (char *)"RAR";
        if (memcmp(buf, "MZ", 2) == 0)                  return (char *)"EXE";
        if (memcmp(buf, "\000\000\001\263", 4) == 0)    return (char *)"MPEG";
        if (memcmp(buf, "MOVI", 4) == 0)                return (char *)"MOVI";
        if (memcmp(buf, "\007\007\007", 3) == 0)        return (char *)"CPIO";
        if (memcmp(buf, "\351,\001JAM", 6) == 0)        return (char *)"JAM";
        if (memcmp(buf, "SQSH", 4) == 0)                return (char *)"SQSH";
        if (memcmp(buf, "UC2\0x1a", 4) == 0)            return (char *)"UC2";
        if (memcmp(buf, ".snd", 4) == 0)                return (char *)"SND";
        if (memcmp(buf, "MThd", 4) == 0)                return (char *)"MID";
        if (memcmp(buf, "RIFF", 4) == 0)                return (char *)"WAV";
        if (memcmp(buf, "EMOD", 4) == 0)                return (char *)"MOD";
        if (memcmp(buf, "MTM", 3) == 0)                 return (char *)"MTM";
        if (memcmp(buf, "#/bin/", 6) == 0)              return (char *)"UNIX script";
        if (memcmp(buf, "\037\235", 2) == 0)            return (char *)"Compressed data";
        if (memcmp(buf, "\037\213", 2) == 0)            return (char *)"gzip compress";
        if (memcmp(buf, "\177ELF", 4) == 0)             return (char *)"ELF";
        if (memcmp(buf, "%!", 2) == 0)                  return (char *)"PostScript";
        if (memcmp(buf, "GIF8", 4) == 0)                return (char *)"GIF";
        if (memcmp(buf, "\377\330\377\340", 4) == 0)    return (char *)"JPEG";
        if (memcmp(buf, "\377\330\377\356", 4) == 0)    return (char *)"JPG";
        if (memcmp(buf, "BM", 2) == 0)                  return (char *)"Bitmap";
        if (memcmp(buf, "%PDF", 4) == 0)                return (char *)"PDF";
        if (memcmp(buf, "THNL", 4) == 0)                return (char *)"ThumbNail";
        if ((memcmp(buf, "<html>", 6) == 0) ||
            (memcmp(buf, "<HTML>", 6) == 0))            return (char *)"HTML";

        /*
         * .COM formats. Should cover about 2/3 of COM files.
         */
        if ((*buf == 0xe9) || (*buf == 0x8c) ||
            (*buf == 0xeb) || (*buf == 0xb8))           return (char *)"COM";

        return NULL;
}
